from typing import Callable, Optional

from bottleneck import allnan as _allnan
from bottleneck import move_argmax as _move_argmax
from bottleneck import move_argmin as _move_argmin
from bottleneck import move_max as _move_max
from bottleneck import move_mean as _move_mean
from bottleneck import move_median as _move_median
from bottleneck import move_min as _move_min
from bottleneck import move_rank as _move_rank
from bottleneck import move_std as _move_std
from bottleneck import move_sum as _move_sum
from bottleneck import move_var as _move_var
from bottleneck import nanmax as _nanmax
from bottleneck import nanmean as _nanmean
from bottleneck import nanmedian as _nanmedian
from bottleneck import nanmin as _nanmin
from bottleneck import nanstd as _nanstd
from bottleneck import nansum as _nansum
from numpy import abs as _abs
from numpy import asarray as _asarray
from numpy import concatenate as _concatenate
from numpy import exp as _exp
from numpy import float64
from numpy import full as _full
from numpy import isnan as _isnan
from numpy import log as _log
from numpy import maximum as _maximum
from numpy import nan as _nan
from numpy import nancumsum as _nancumsum
from numpy import nanpercentile as _nanpercentile
from numpy import sqrt as _sqrt
from numpy import zeros as _zeros
from numpy.lib.stride_tricks import sliding_window_view as _sliding_window_view
from numpy.typing import NDArray
from scipy.optimize import minimize as _minimize
from scipy.stats import kurtosis as _kurtosis
from scipy.stats import skew as _skew

from tompy.scalar import div as _div
from tompy.scalar import lr2sr as _lr2sr


def __sr2lr(sr: NDArray[float64], /) -> NDArray[float64]:
    return _log(sr + 1)


def __lr2sr(lr: NDArray[float64], /) -> NDArray[float64]:
    return _exp(lr) - 1  # type: ignore


def validate(v: NDArray[float64], /) -> int:
    assert v.ndim == 1
    return v.shape[0]


def validate_rolling(v: NDArray[float64], /, window: int) -> int:
    assert window >= 0
    return validate(v)


def zeros(n: int, /) -> NDArray[float64]:
    return _zeros(n, dtype=float64)


def nans(n: int, /) -> NDArray[float64]:
    return _full(n, _nan, dtype=float64)


def describe_meta() -> dict[str, str]:
    return {
        "COUNT": "Count",
        "MEAN": "Mean",
        "STD": "Standard Deviation",
        "SKEW": "Skewness",
        "KURT": "Kurtosis",
        "MIN": "Min",
        "25%": "25%",
        "50%": "50%",
        "75%": "75%",
        "MAX": "Max",
    }


def describe(v: NDArray[float64], /) -> NDArray[float64]:
    n = validate(v)
    if n <= 0:
        return nans(len(describe_meta()))
    return _asarray(
        [
            nancnt(v),
            nanmean(v),
            nanstd(v),
            nanskew(v),
            nankurt(v),
            nanpctl(v, 0),
            nanpctl(v, 25),
            nanpctl(v, 50),
            nanpctl(v, 75),
            nanpctl(v, 100),
        ],
        dtype=float64,
    )


def shift(v: NDArray[float64], /, lag: int) -> NDArray[float64]:
    n = validate(v)
    if n <= 0 or lag == 0:
        return v
    if lag > 0:
        if lag >= n:
            return nans(n)
        return _concatenate((nans(lag), v[:-lag]))
    lag = -lag
    if lag >= n:
        return nans(n)
    return _concatenate((v[lag:], nans(lag)))


def nancnt(v: NDArray[float64], /) -> float:
    n = validate(v)
    if n <= 0:
        return _nan
    return n - _nansum(_isnan(v))  # type: ignore


def nanmean(v: NDArray[float64], /) -> float:
    n = validate(v)
    if n <= 0:
        return _nan
    return _nanmean(v)  # type: ignore


def nanstd(v: NDArray[float64], /) -> float:
    n = validate(v)
    if n <= 0:
        return _nan
    return _nanstd(v, ddof=1)  # type: ignore


def nanskew(v: NDArray[float64], /) -> float:
    n = validate(v)
    if n <= 0:
        return _nan
    if _allnan(v):
        return _nan
    return _skew(v, axis=None, bias=False, nan_policy="omit")  # type: ignore


def nankurt(v: NDArray[float64], /) -> float:
    n = validate(v)
    if n <= 0:
        return _nan
    if _allnan(v):
        return _nan
    return _kurtosis(  # type: ignore
        v, axis=None, fisher=True, bias=False, nan_policy="omit"
    )


def nanpctl(v: NDArray[float64], /, q: int) -> float:
    n = validate(v)
    if n <= 0:
        return _nan
    if q <= 0:
        return _nanmin(v)  # type: ignore
    if q >= 100:
        return _nanmax(v)  # type: ignore
    if q == 50:
        return _nanmedian(v)  # type: ignore
    if _allnan(v):
        return _nan
    return _nanpercentile(v, q)  # type: ignore


def rolling_sum(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_sum(v, window)  # type: ignore


def rolling_mean(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_mean(v, window)  # type: ignore


def rolling_std(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_std(v, window, ddof=1)  # type: ignore


def rolling_var(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_var(v, window, ddof=1)  # type: ignore


def rolling_min(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_min(v, window)  # type: ignore


def rolling_max(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_max(v, window)  # type: ignore


def rolling_argmin(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_argmin(v, window)  # type: ignore


def rolling_argmax(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_argmax(v, window)  # type: ignore


def rolling_median(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_median(v, window)  # type: ignore


def rolling_rank(v: NDArray[float64], /, window: int) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    return _move_rank(v, window)  # type: ignore


def rolling_apply(
    v: NDArray[float64],
    /,
    window: int,
    func: Callable[[NDArray[float64]], float],
) -> NDArray[float64]:
    n = validate_rolling(v, window)
    if window == 0 or window > n:
        return nans(n)
    swv = _sliding_window_view(v, window)
    ret = _asarray([func(sw) for sw in swv], dtype=float64)
    assert ret.ndim == 1
    if window > 1:
        ret = _concatenate((nans(window - 1), ret))
    return ret


def price2relative(price: NDArray[float64], /) -> NDArray[float64]:
    n = validate(price)
    if n <= 1:
        return nans(n)
    rel = price[1:] / price[:-1]
    return _concatenate((nans(1), rel))


def price2sr(price: NDArray[float64], /) -> NDArray[float64]:
    n = validate(price)
    if n <= 1:
        return nans(n)
    sr = price[1:] / price[:-1] - 1
    return _concatenate((nans(1), sr))


def price2lr(price: NDArray[float64], /) -> NDArray[float64]:
    n = validate(price)
    if n <= 1:
        return nans(n)
    lr = _log(price[1:] / price[:-1])
    return _concatenate((nans(1), lr))


def price_metrics(
    price: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    rel = price2relative(price)
    sr = rel - 1
    lr = _log(rel)
    return lr_sr_metrics(
        lr, sr, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
    )


def price_terms_metrics(
    price: NDArray[float64],
    terms: list[int],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    n = validate(price)
    rel = price2relative(price)
    sr = rel - 1
    lr = _log(rel)
    if sqrt_ann_factor is None:
        sqrt_ann_factor = _sqrt(ann_factor)
    nr = len(lr_sr_metrics_meta(mode))

    def metrics(term: int) -> NDArray[float64]:
        if term > n:
            return nans(nr)
        return lr_sr_metrics(
            lr[-term:],
            sr[-term:],
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        )

    return _asarray([metrics(term) for term in terms], dtype=float64).T


def sr2lr(sr: NDArray[float64], /) -> NDArray[float64]:
    n = validate(sr)
    if n <= 0:
        return nans(n)
    return __sr2lr(sr)


def sr_kelly_criterion(sr: NDArray[float64], /, ann_factor: float) -> float:
    """
    Kelly Criterion with Empirical Distribution
    """
    n = validate(sr)
    if n <= 0:
        return _nan

    min_sr = _nanmin(sr)
    max_sr = _nanmax(sr)
    if max_sr <= 0 or min_sr >= 0:
        return _nan

    eps = 0.000001
    lb = -1 / max_sr
    ub = -1 / min_sr
    bounds = [(lb + eps, ub - eps)]
    x0 = zeros(1)
    result = _minimize(
        lambda x: -_nanmean(_log(sr * x[0] + 1)) * ann_factor,
        x0,
        method="SLSQP",
        bounds=bounds,
    )
    if result.success:
        return result.x[0]  # type: ignore
    return _nan


def sr_metrics(
    sr: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return lr_sr_metrics(
        sr2lr(sr), sr, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
    )


def lr2sr(lr: NDArray[float64], /) -> NDArray[float64]:
    n = validate(lr)
    if n <= 0:
        return nans(n)
    return __lr2sr(lr)


def lr2crdd(
    lr: NDArray[float64], /
) -> tuple[NDArray[float64], NDArray[float64]]:
    n = validate(lr)
    if n <= 0:
        return nans(n), nans(n)
    lr0_cr = _nancumsum(lr)
    lr0_peak = _maximum.accumulate(lr0_cr)
    lr0_dd = lr0_cr - lr0_peak
    return lr0_cr, lr0_dd


def lr_metrics(
    lr: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return lr_sr_metrics(
        lr, lr2sr(lr), ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
    )


def lr_sr_metrics_meta(mode: int, /) -> dict[str, str]:
    if mode == 0:
        return {
            # Basic
            "SR_CR": "[SR] Cumulative Return = P(T) / P(0) - 1",
            "SR_AGR": "[SR] Annualized Growth Rate (aka. CAGR)",
            "SR_VOL": "[SR] Annualized Volatility",
            "SR_MDD": "[SR] Maximum Drawdown = min(SR_DD)",
            # Risk Adjusted Return
            "SR_SHARPE": "[SR] Sharpe Ratio = SR_AGR / SR_VOL",
            "SR_SORTINO": "[SR] Sortino Ratio = SR_AGR / SR_DVOL",
            "SR_CALMAR": "[SR] Calmar Ratio = SR_AGR / -SR_MDD",
        }
    if mode == 1:
        return {
            # Basic
            "LR_CR": "[LR] Cumulative Return = log(P(T) / P(0))",
            "LR_AGR": "[LR] Annualized Growth Rate",
            "LR_VOL": "[LR] Annualized Volatility",
            "LR_MDD": "[LR] Maximum Drawdown = min(LR_DD)",
            # Risk Adjusted Return
            "LR_SHARPE": "[LR] Sharpe Ratio = LR_AGR / LR_VOL",
            "LR_SORTINO": "[LR] Sortino Ratio = LR_AGR / LR_DVOL",
            "LR_CALMAR": "[LR] Calmar Ratio = LR_AGR / -LR_MDD",
        }
    return {
        # Basic
        "LR_CR": "[LR] Cumulative Return = log(P(T) / P(0))",
        "LR_AGR": "[LR] Annualized Growth Rate",
        "LR_VOL": "[LR] Annualized Volatility",
        "LR_MDD": "[LR] Maximum Drawdown = min(LR_DD)",
        # Risk Adjusted Return
        "LR_SHARPE": "[LR] Sharpe Ratio = LR_AGR / LR_VOL",
        "LR_SORTINO": "[LR] Sortino Ratio = LR_AGR / LR_DVOL",
        "LR_CALMAR": "[LR] Calmar Ratio = LR_AGR / -LR_MDD",
        "SR_KELLY": "[SR] Kelly Criterion",
        # Potential
        "LR_PAGR": "[LR] Potential Annualized Growth Rate",
        "LR_ARG/PAGR": "[LR] LR_ARG / LR_PAGR",
        # Ratio
        "LR_WINS/TOTALS": "[LR] Wins/Totals Ratio = LR_P_COUNT / LR_COUNT",
        "LR_LOSSES/WINS": "[LR] Losses/Wins Ratio = LR_L_COUNT / LR_P_COUNT",
        "LR_PROFIT/LOSS": "[LR] Profit/Loss Ratio = LR_P_MEAN / -LR_L_MEAN",
        "LR_PROFIT_FACTOR": "[LR] Profit Factor = LR_P_SUM / -LR_L_SUM",
        # Etc
        "LR_DVOL": "[LR] Annualized Downside Volatility",
        "LR_DD": "[LR] Drawdown = LR_CR - LR_PEAK",
        # Descriptive Statistics
        "LR_COUNT": "[LR] Count",
        "LR_MEAN": "[LR] Mean",
        "LR_STD": "[LR] Standard Deviation",
        "LR_SKEW": "[LR] Skewness",
        "LR_KURT": "[LR] Kurtosis",
        "LR_MIN": "[LR] Min",
        "LR_25%": "[LR] 25%",
        "LR_50%": "[LR] 50%",
        "LR_75%": "[LR] 75%",
        "LR_MAX": "[LR] Max",
    }


def lr_sr_metrics(
    lr: NDArray[float64],
    sr: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    n_lr = validate(lr)
    n_sr = validate(sr)
    assert n_lr == n_sr
    if n_lr <= 0:
        return nans(len(lr_sr_metrics_meta(mode)))
    if sqrt_ann_factor is None:
        sqrt_ann_factor = _sqrt(ann_factor)

    lr0_cr, lr0_dd = lr2crdd(lr)
    lr_cr = lr0_cr[-1]
    lr_mdd = _nanmin(lr0_dd)
    lr_mean = nanmean(lr)
    lr_agr = lr_mean * ann_factor
    if mode == 0:
        srl = sr[sr <= 0]
        sr_cr = _lr2sr(lr_cr)
        sr_agr = _lr2sr(lr_agr)
        sr_vol = nanstd(sr) * sqrt_ann_factor
        sr_dvol = nanstd(srl) * sqrt_ann_factor
        sr_mdd = _lr2sr(lr_mdd)
        sr_sharpe = _div(sr_agr, sr_vol)
        sr_sortino = _div(sr_agr, sr_dvol)
        sr_calmar = _div(sr_agr, -sr_mdd)
        return _asarray(
            [sr_cr, sr_agr, sr_vol, sr_mdd, sr_sharpe, sr_sortino, sr_calmar],
            dtype=float64,
        )

    lr_std = nanstd(lr)
    lr_vol = lr_std * sqrt_ann_factor
    lrl = lr[lr <= 0]
    lr_dvol = nanstd(lrl) * sqrt_ann_factor
    lr_sharpe = _div(lr_agr, lr_vol)
    lr_sortino = _div(lr_agr, lr_dvol)
    lr_calmar = _div(lr_agr, -lr_mdd)
    if mode == 1:
        return _asarray(
            [lr_cr, lr_agr, lr_vol, lr_mdd, lr_sharpe, lr_sortino, lr_calmar],
            dtype=float64,
        )

    lr_count = nancnt(lr)
    lrp = lr[lr > 0]
    lrp_count = nancnt(lrp)
    lrp_mean = nanmean(lrp)
    lrl_count = nancnt(lrl)
    lrl_mean = nanmean(lrl)
    lr_pagr = nanmean(_abs(lr)) * ann_factor
    return _asarray(
        [
            lr_cr,
            lr_agr,
            lr_vol,
            lr_mdd,
            lr_sharpe,
            lr_sortino,
            lr_calmar,
            sr_kelly_criterion(sr, ann_factor),
            lr_pagr,
            _div(lr_agr, lr_pagr),
            _div(lrp_count, lr_count),
            _div(lrl_count, lrp_count),
            _div(lrp_mean, -lrl_mean),
            _div(lrp_mean * lrp_count, -lrl_mean * lrl_count),
            lr_dvol,
            lr0_dd[-1],
            lr_count,
            lr_mean,
            lr_std,
            nanskew(lr),
            nankurt(lr),
            nanpctl(lr, 0),
            nanpctl(lr, 25),
            nanpctl(lr, 50),
            nanpctl(lr, 75),
            nanpctl(lr, 100),
        ],
        dtype=float64,
    )
